#!/usr/bin/perl
#
# WardenWatchdog.pm
#
# Copyright (C) 2011-2012 Cesnet z.s.p.o
#
# Use of this source is governed by a BSD-style license, see LICENSE file.

package WardenWatchdog;

#use Data::Dumper;
use WardenConf;
use strict;
use warnings;
use DBI;
use DBD::mysql;
use DateTime;
use Email::Simple;
use Sys::Hostname;

#-------------------------------------------------------------------------------
# sendmail_wrapper
#
# Simple wrapper function for an mailserver.
#
# Input:
#     message    = prepared message
#     email_conf = configuration of a mailserver
#
# Output: -
#
# Return:
#   On Success (1)
#   On Failure (0, 'Error message')
#-------------------------------------------------------------------------------
sub sendmail_wrapper{

  my $message    = shift;
  my $email_conf = shift;

  if(open(my $sendmail, $email_conf)){
    print $sendmail $message;
    close $sendmail;
    return (1);
  } else {
    return (0, "Sending email failed: $!");
  }
}

# Array of hashes
#{query => ; text => ; contact => }


#-------------------------------------------------------------------------------
# send_report
#
# Function for creating and sending of an Watchdog status report via email to
# administrators of an clients.
#
# Input:
#   Hash of parameters:
#     contact    => email address of a message recipient
#     domain     => domain name of server where script runs
#     subject    => subject of an email
#     text       => body of an email
#     email_conf => configuration of a mailserver
#
# Output: -
#
# Return:
#   On Success (1)
#   On Failure (0, 'Error message')
#-------------------------------------------------------------------------------
sub send_report{

  my $input_data = shift;
  my $contact    = $$input_data{'contact'};
  my $domain     = $$input_data{'domain'};
  my $subject    = $$input_data{'subject'};
  my $text       = $$input_data{'text'};
  my $email_conf = $$input_data{'email_conf'};

  my $message;
  if(!($contact)){
    return (0, "Empty 'To' email header!\n");
  }

  if(!($domain)){
    return (0, "No sender's domain! Can't send email\n");
  }

  if(!($text)){
    return (0, "No text! Nothing to send\n");
  }

  eval{
    $message = Email::Simple->create(
    header => [
      To => $contact,
      From => 'warden_watchdog@'.$domain,
      Subject => $subject],
    body => $text);
  } or do {
    return (0, "Can't create email message\n");
  };

  my ($rc, $err) = sendmail_wrapper($message->as_string,$email_conf);
  if(!$rc){
    return (0, $err);
  }
  return (1);
}

#-------------------------------------------------------------------------------
# connect_to_DB
#
# Just simple database wrapper for Watchdog which creates db's handler.
#
# Input:
#   Hash of parameters:
#     platform => database platform name according to Perl DBI
#     name     => name of a database
#     hostname => database hostname
#     user     => username
#     passwd   => password
#
# Output:
#   dbhRef = reference on a database handler
#
# Return:
#   On Success (1)
#   On Failure (0, 'Error message')
#-------------------------------------------------------------------------------
sub connect_to_DB {

  my $dbConf = shift;
  my $dbhRef = shift;

  my $dbPlatform = $$dbConf{'platform'};
  my $dbName     = $$dbConf{'name'};
  my $dbHostname = $$dbConf{'hostname'};
  my $dbUser     = $$dbConf{'user'};
  my $dbPasswd   = $$dbConf{'passwd'};

  my $dbh;
  if($dbh = DBI->connect( "dbi:$dbPlatform:database=$dbName;host=$dbHostname", $dbUser, $dbPasswd, {mysql_auto_reconnect => 1})){
    $$dbhRef = $dbh;
    return (1);
  }
  else{
    return (0,"Cannot connect to database! ".DBI->errstr);
  }
}

#-------------------------------------------------------------------------------
# update_procedures
#
# Function takes DB handler and executes all database procedures in the array.
#
# Input:
#   dbhRef  = reference on a database handler
#   procRef = reference on an array of database procedures
#
# Output: -
#
# Return:
#   On Success (1)
#   On Failure (0, 'Error message')
#-------------------------------------------------------------------------------
sub update_procedures{

  my $dbhRef  = shift;
  my $procRef = shift;

  my $dbh = $$dbhRef;
  if(!defined($dbh)){
    return (0, "update_procedures: no database handle defined")
  }

  my @sqlQueries = @{$procRef};
  foreach my $proc (@sqlQueries) {
    $dbh->do($proc);
  }

  return (1);
}

#-------------------------------------------------------------------------------
# send_query
#
#
#
# Input:
#   dbhRef    = reference on a database handler
#   configRef = Hash of parameters:
#     query   => sql query of an action (check) on Warden database
#     text    => body of an email which is send to a admin of an client
#                in case of nonempty check result
#     contact => contact for message, which overrides contact collumn
#                in a database table.
#
# Output:
#   eventsRef = Hash of parameters:
#     contact   = email address of an client administrator
#     'contact' => predefined email text + information from database obtained
#                  by a query
#
# Return:
#   On Success (1)
#   On Failure (0, 'Error message')
#-------------------------------------------------------------------------------
sub send_query{

  my $dbhRef    = shift;
  my $configRef = shift;
  my $eventsRef = shift;

  my $dbh = $$dbhRef;
  if(!defined($dbh)){
    return (0, "send_query: no database handle defined")
  }

  my @config = @{$configRef};
  my %bad_events;
  my ($rc,$err);
  my $i = 0;

  while ($i < scalar(@config)) {
    # run DB query -> requestor, client name
    my $sth;
    if (defined($config[$i]{query})){
      $sth = $dbh->prepare($config[$i]{query});
    }
    else{
      return (0, "No query available\n");
    }

    if (!($sth->execute)){
      return (0, "Couldn't get data from my database: $sth->errstr\n");
    };

    my $contact;
    my $msg_text = 1;

    while(my $result = $sth->fetchrow_hashref()){
      if (defined($config[$i]{contact})){
        # override contact from 'requestor' collumn
        $contact = $config[$i]{contact};
      }
      else{
        $contact = $result->{'requestor'};
      }
      # information header
      if($msg_text){
        $bad_events{$contact} .= $config[$i]{text} . "\n\n";
        $bad_events{$contact} .= join(" | ", map {$_ // "UNKNOWN" } keys %$result) . "\n";
        $msg_text = 0;
      }
      $bad_events{$contact} .= join(" | ", map {$_ // "NULL" } values %$result) . "\n";
    }
    foreach my $key (keys %bad_events){
      $bad_events{$key} .= "\n\n";
    }

    $sth->finish;
    $i++;
  }

  %$eventsRef = %bad_events;
  return (1);
}

#-------------------------------------------------------------------------------
# run
#
# Main module function which takes configure file, interval and performs Warden
# database checks described in a configuration file. Results are send to client
# administrators or on a defined addresses if they are set in a config.
#
# Input:
#   conf_file = path to a config file with database procedures and other params
#   period    = interval in days from now back to the past for database check
#
# Output: -
#
# Return:
#   On Success (1)
#   On Failure (0, 'Error message')
#-------------------------------------------------------------------------------
sub run{

  my $conf_file = shift;
  my $period    = shift;

  my $errMsg;
  # server config

  if(!defined($conf_file)){
    return (0,"No conf file is available");
  }

  if(!defined($period)){
    return (0,"No time period is defined");
  }

  our $server_conf       = undef;
  our $domain_name       = undef;
  our $email_subject     = undef;
  our $email_server_conf = undef;
  our @sql_precondition  = undef;
  our @sql_queries       = undef;
  our @sql_postcondition = undef;

  # script config
  if (!(do $conf_file)) {
    if ($@){
      $errMsg = "Errors in config file '$conf_file': $@";
      #syslog("err|$errMsg");
      print $errMsg;
      return (0,"Warden watchdog - $errMsg");
    }
    if (!(defined $_)){
      $errMsg = "Can't read config file '$conf_file': $!";
      #syslog("err|$errMsg");
      print $errMsg;
      return (0,"Warden watchdog - $errMsg");
    }
  }

  # server config
  our $SYSLOG_VERBOSE  = undef;
  our $SYSLOG_FACILITY = undef;
  our $DB_NAME         = undef;
  our $DB_USER         = undef;
  our $DB_PASS         = undef;
  our $DB_HOST         = undef;

  # TODO replace with function call from Wardencommon
  if (!(do $server_conf)) {
    if ($@){
      $errMsg = "Errors in config file '$server_conf': $@";
      #syslog("err|$errMsg");
      print $errMsg;
      return (0,"Warden watchdog - $errMsg");
    }
    if (!(defined $_)){
      $errMsg = "Can't read config file '$server_conf': $!";
      #syslog("err|$errMsg");
      print $errMsg;
      return (0,"Warden watchdog - $errMsg");
    }
  }

  #print "SYSLOG_VERBOSE $SYSLOG_VERBOSE, SYSLOG_FACILITY $SYSLOG_FACILITY, DB_NAME $DB_NAME, DB_USER $DB_USER, DB_PASS $DB_PASS, DB_HOST $DB_HOST";
  my %db_conf = (platform => 'mysql', name => $DB_NAME, hostname => $DB_HOST, user => $DB_USER, passwd => $DB_PASS);

  my $date;
  eval{
    my $dt = DateTime->now(time_zone => 'UTC')->subtract(days => $period);
    $dt->set_time_zone('local');
    $date = $dt->strftime("%Y-%m-%d %H:%M:%S");
  } or do {
    #print "Warden watchdog - can't work with date\n";
    syslog("Warden watchdog - can't work with date\n");
  };



  my $dbh;
  # connect to DB
  my ($rc,$err) = connect_to_DB(\%db_conf,\$dbh);
  if (!$rc){
    $errMsg = "Warden watchdog can\'t connect do DB: $err";
    return (0,"Warden watchdog - $errMsg");
  }

  if(@sql_precondition){
    ($rc,$err) = update_procedures(\$dbh,\@sql_precondition);
    if (!$rc){
      #print "Warden watchdog - $err\n";
      return (0,"Warden watchdog - $err\n");
    }
  }

  my %bad_events;
  if(@sql_queries){
    foreach my $query (@sql_queries){
      $query->{query} =~ s/\$date/$date/;
    }
    my ($rc,$err) = send_query(\$dbh,\@sql_queries,\%bad_events);
    if (!$rc){
      #print "Warden watchdog - $err\n";
      return (0,"Warden watchdog - $err\n");
    }
  }

  if(@sql_postcondition){
    my ($rc,$err) = update_procedures(\$dbh,\@sql_postcondition);
    if (!$rc){
      #print "Warden watchdog - $err\n";
      return (0,"Warden watchdog - $err\n");
    }
  }

  while (my ($contact, $text) = each(%bad_events)){
    my %input = (contact => $contact, domain => $domain_name, text => $text, subject => $email_subject, email_conf => $email_server_conf);
    my ($rc,$err) = send_report(\%input);
    if (!$rc){
      #print $err;
      return (0,"Warden client - networkReporter $err\n");
    }
  }

  # disconnect to DB
  $dbh->disconnect;
}
1;
